import os
import json
import debiaiServer.utils.utils as utils
import uuid
from debiaiServer.debiai_gui_utils import data_folder_path

CONF_PATH = data_folder_path + "/widgetConfigurations.json"

# Configuration file structure
# {
#     "widgetKey": [
#         {
#             "id": ""
#             "name": ""
#             "description": "",
#             "projectId": ""
#             "dataProviderId": ""
#             "creationDate": 0
#
#             "configuration": {},
#         },
#         ...
#      ]
# }


def setup_widget_configurations():
    # Create the file if it does not exist
    if not os.path.exists(CONF_PATH):
        _save_configurations({})


def get_configurations_overview():
    # Return the number of configurations for each widget
    all_configurations = _get_all_configurations()

    configurations_overview = {}
    for widget_key in all_configurations:
        configurations_overview[widget_key] = len(all_configurations[widget_key])

    return configurations_overview


def get_configurations(widget_key):
    # Return the configurations list of the widget
    all_configurations = _get_all_configurations()

    if widget_key in all_configurations:
        return all_configurations[widget_key]
    else:
        return []


def add_configuration(widget_key, data):
    # project_id, data_provider_id, conf_description, conf_name, conf
    # Add a new widget configuration
    configurations = _get_all_configurations()

    if widget_key not in configurations:
        configurations[widget_key] = []

    # Generate id
    id = str(uuid.uuid1())

    configuration_to_add = {
        "id": id,
        "name": data["name"],
        "description": data["description"],
        "projectId": data["projectId"],
        "dataProviderId": data["dataProviderId"],
        "creationDate": utils.timeNow(),
        "configuration": data["configuration"],
    }

    # Save configuration
    configurations[widget_key].append(configuration_to_add)
    _save_configurations(configurations)


def delete_configuration(widget_key, id):
    # Delete the widget configuration by its name
    configurations = _get_all_configurations()

    if widget_key in configurations:
        for configuration in configurations[widget_key]:
            if configuration["id"] == id:
                configurations[widget_key].remove(configuration)

        _save_configurations(configurations)


def _get_all_configurations():
    # Return the configurations list of all widgets
    try:
        with open(CONF_PATH) as json_file:
            return json.load(json_file)

    except FileNotFoundError:
        setup_widget_configurations()
        return {}
    except json.decoder.JSONDecodeError as e:
        print("Error while reading the widget configurations file")
        print(e)
        print("The file will be reset")
        _save_configurations({})
        return {}


def _save_configurations(conf, retry=False):
    # Update the json file
    try:
        with open(CONF_PATH, "w") as json_file:
            json.dump(conf, json_file)
    except FileNotFoundError:
        if not retry:
            setup_widget_configurations()
            _save_configurations(conf, True)
        else:
            print("Error while saving the widget configurations file")
            print("The file will not be saved")



import debiaiServer.modules.dataProviders.webDataProvider.http.api as api
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)


def get_models_info(url, project_id):
    # Models
    try:
        models = api.get_models(url, project_id)
        debiai_models = []
        for model_in in models:
            if "id" not in model_in:
                continue
            model = {
                "id": model_in["id"],
                "metadata": None,
                "creationDate": None,
            }

            # Adding name and nbResults
            model["name"] = model_in["name"] if "name" in model_in else model_in["id"]
            if "nbResults" in model_in:
                model["nbResults"] = model_in["nbResults"]

            # Adding metadata
            if "metadata" in model_in:
                model["metadata"] = model_in["metadata"]

            # Adding creationDate
            if "creationDate" in model_in:
                model["creationDate"] = model_in["creationDate"]

            debiai_models.append(model)

        return debiai_models
    except DataProviderException:
        # The route may not be implemented in the data provider
        return []


def get_model_result_id(url, cache, project_id, model_id):
    # Todo : Add route to call Id results for a Model (DP)
    # Todo : Add Some formatting if data has to change

    id_list = cache.get_model_result_id_list(project_id, model_id)

    if id_list is None:
        id_list = api.get_model_result_id_list(url, project_id, model_id)
        cache.set_model_result_id_list(project_id, model_id, id_list)

    return id_list


def get_model_results(url, project_id, model_id, sample_list):
    return api.get_model_result(url, project_id, model_id, sample_list)


def delete_model(url, project_id, model_id):
    return api.delete_model(url, project_id, model_id)



from debiaiServer.config.init_config import get_config
import debiaiServer.modules.exportMethods.exportUtils as exportUtils

#############################################################################
# Export API Management
#############################################################################


def get_export_methods():
    # ParametersCheck
    return exportUtils.get_export_methods(), 200


def post_export_method(data):
    # Check if the creation of export methods is allowed
    config = get_config()
    creation_allowed = config["EXPORT_METHODS_CONFIG"]["creation"]
    if not creation_allowed:
        return "Export method creation is not allowed", 403

    try:
        return exportUtils.add_export_method(data), 200
    except Exception as e:
        return str(e), 400


def delete_export_method(exportMethodId):
    # Check if the deletion of export methods is allowed
    config = get_config()
    deletion_allowed = config["EXPORT_METHODS_CONFIG"]["deletion"]
    if not deletion_allowed:
        return "Export method deletion is not allowed", 403

    try:
        return exportUtils.delete_export_method(exportMethodId), 200
    except Exception as e:
        return str(e), 400


def exportSelection(dataProviderId, projectId, data):
    try:
        return exportUtils.exportSelection(dataProviderId, projectId, data), 200
    except Exception as e:
        return str(e), 400


def exportData(exportMethodId, data):
    try:
        return exportUtils.exportData(exportMethodId, data), 200
    except Exception as e:
        return str(e), 400



import os
import shutil
import ujson as json

from debiaiServer.modules.dataProviders.pythonDataProvider.dataUtils import (
    pythonModuleUtils,
    hash,
)

DATA_PATH = pythonModuleUtils.DATA_PATH


def project_exist(projectId):
    return projectId in os.listdir(DATA_PATH)


def _get_project_info(projectId):
    file_path = DATA_PATH + projectId + "/info.json"
    if not os.path.exists(file_path):
        delete_project(projectId)
        raise Exception("The project info file is missing")

    with open(file_path) as json_file:
        data = json.load(json_file)

    if "name" not in data:
        delete_project(projectId)
        raise Exception("The project name is missing from the info.json file")
    if "creationDate" not in data:
        delete_project(projectId)
        raise Exception("The project creationDate is missing from the info.json file")
    if "updateDate" not in data:
        delete_project(projectId)
        raise Exception("The project updateDate is missing from the info.json file")

    return data


def get_project(projectId):
    try:
        # Get info file
        data = _get_project_info(projectId)
        name = data["name"]
        creationDate = data["creationDate"]
        updateDate = data["updateDate"]

        # Nb models
        if not os.path.exists(DATA_PATH + projectId + "/models/"):
            raise Exception('The "models" folder is missing')

        nbModels = len(os.listdir(DATA_PATH + projectId + "/models/"))

        # Nb selection
        if not os.path.exists(DATA_PATH + projectId + "/selections/"):
            raise Exception('The "selections" folder is missing')

        nbSelection = len(os.listdir(DATA_PATH + projectId + "/selections/"))

        # Nb samples
        if not os.path.exists(DATA_PATH + projectId + "/samplesHashmap.json"):
            raise Exception('The "samplesHashmap.json" file is missing')

        nbSamples = len(hash.getHashmap(projectId))

        # project columns
        projectColumns = get_project_columns(projectId)

        # project block level
        # We still need to get the project block level, the Python module use it
        projectBlockLevel = get_project_block_level_info(projectId)

        projectOverview = {
            "id": projectId,
            "name": name,
            "nbModels": nbModels,
            "nbSelections": nbSelection,
            "nbSamples": nbSamples,
            "creationDate": creationDate,
            "updateDate": updateDate,
            "columns": projectColumns,
            "blockLevelInfo": projectBlockLevel,
        }

    except Exception as e:
        print("Error while getting the project overview: " + projectId)
        print(e)
        projectOverview = {
            "id": projectId,
            "name": projectId,
        }

    return projectOverview


def get_projects():
    project = []

    for projectId in os.listdir(DATA_PATH):
        project.append(get_project(projectId))

    return project


def create_project(projectId, projectName):
    # Create the project files and folders
    os.mkdir(DATA_PATH + projectId)
    os.mkdir(DATA_PATH + projectId + "/blocks")
    os.mkdir(DATA_PATH + projectId + "/models")
    os.mkdir(DATA_PATH + projectId + "/selections")

    now = pythonModuleUtils.timeNow()
    projectInfo = {
        "name": projectName,
        "id": projectId,
        "creationDate": now,
        "updateDate": now,
        "blockLevelInfo": [],
    }

    pythonModuleUtils.writeJsonFile(DATA_PATH + projectId + "/info.json", projectInfo)
    pythonModuleUtils.writeJsonFile(DATA_PATH + projectId + "/samplesHashmap.json", {})

    return projectInfo


def update_project(projectId):
    # Change the update date of the project to now
    pythonModuleUtils.updateJsonFile(
        DATA_PATH + projectId + "/info.json", "updateDate", pythonModuleUtils.timeNow()
    )


def get_project_block_level_info(projectId):
    data = _get_project_info(projectId)
    if "blockLevelInfo" in data:
        return data["blockLevelInfo"]


def get_project_columns(projectId):
    block_level_info = get_project_block_level_info(projectId)

    # Convert the block level info to the new columns format
    #   blockLevelInfo:
    #   [
    #     { "name": "block1" },
    #     {
    #       "name": "block2",
    #       "contexts": [
    #           { "name": "cont1", "type": "text", group:"group_1" },
    #           { "name": "cont2", "type": "text", group:"group_1" },
    #       ]
    #     },
    #     { "name": "block3", "contexts": [
    #           { "name": "cont3", "type": "text", group:"group_1" }
    #        ]
    #     },
    #     {
    #       "name": "block4",
    #       "others": [{ "name": "other1", "type": "number" }],
    #       "groundTruth": [
    #           { "name": "gdt1", "type": "number" },
    #           { "name": "gdt2", "type": "number" },
    #       ],
    #       "inputs": [
    #           { "name": "inp1", "type": "number" }
    #       ]
    #     }
    # ]

    # Goal format:
    # [
    #   { "name": "block1", "category": "other", "type": "auto" },
    #   { "name": "block2", "category": "other", "type": "auto" },
    #   { "name": "cont1", "category": "context", "type": "text", group: "group_1" },
    #   { "name": "cont2", "category": "context", "type": "text", group: "group_1" },
    #   { "name": "cont3", "category": "context", "type": "text", group: "group_1" },
    #   { "name": "block3", "category": "other", "type": "auto" },
    #   { "name": "other1", "category": "other", "type": "number" },
    #   { "name": "block4", "category": "other", "type": "auto" },
    #   { "name": "gdt1", "category": "groundtruth", "type": "number" },
    #   { "name": "gdt2", "category": "groundtruth", "type": "number" },
    #   { "name": "inp1", "category": "input", "type": "number" },
    # ]

    project_columns = []

    def create_column(col, category):
        column = {"name": col["name"], "category": category, "type": col["type"]}

        if "group" in col:
            column["group"] = col["group"]

        return column

    for block in block_level_info:
        block_name = block["name"]
        project_columns.append(
            {"name": block_name, "category": "other", "type": "auto"}
        )

        if "groundTruth" in block:
            for ground_truth in block["groundTruth"]:
                project_columns.append(create_column(ground_truth, "groundtruth"))

        if "contexts" in block:
            for context in block["contexts"]:
                project_columns.append(create_column(context, "context"))

        if "inputs" in block:
            for input in block["inputs"]:
                project_columns.append(create_column(input, "input"))

        if "others" in block:
            for other in block["others"]:
                project_columns.append(create_column(other, "other"))

    return project_columns


def get_result_structure(projectId):
    data = _get_project_info(projectId)
    if "resultStructure" in data:
        return data["resultStructure"]
    else:
        return None


def delete_project(projectId):
    # Delete the project files and folders
    try:
        shutil.rmtree(DATA_PATH + projectId)
    except Exception as e:
        print(e)
        raise "Something went wrong when deleting the project"


def update_block_structure(projectId, blockStructure):
    try:
        pythonModuleUtils.updateJsonFile(
            DATA_PATH + projectId + "/info.json", "blockLevelInfo", blockStructure
        )

        update_project(projectId)
    except Exception as e:
        print(e)
        raise Exception("Something went wrong updating project structure")


def update_results_structure(projectId, resultStructure):
    try:
        # save resultStructure
        pythonModuleUtils.updateJsonFile(
            DATA_PATH + projectId + "/info.json", "resultStructure", resultStructure
        )
        update_project(projectId)
        return resultStructure, 200

    except Exception as e:
        print(e)
        raise "Something went wrong updating project structure"



import os
import ujson as json

from debiaiServer.modules.dataProviders.pythonDataProvider.dataUtils import (
    pythonModuleUtils,
    projects,
)

DATA_PATH = pythonModuleUtils.DATA_PATH


# Selections
def create_selection(project_id, selection_name, sample_ids, request_id=None):
    selection_id = pythonModuleUtils.clean_filename(selection_name)
    if len(selection_id) == 0:
        selection_id = pythonModuleUtils.timeNow()

    nbS = 1
    while selection_exist(project_id, selection_id):
        selection_id = pythonModuleUtils.clean_filename(selection_name) + "_" + str(nbS)
        nbS += 1

    # Save the selection
    selectionInfoFilePath = (
        DATA_PATH + project_id + "/selections/" + selection_id + "/info.json"
    )
    now = pythonModuleUtils.timeNow()

    selectionInfo = {
        "id": selection_id,
        "name": selection_name,
        "filePath": selectionInfoFilePath,
        "creationDate": now,
        "updateDate": now,
        "samples": sample_ids,
    }

    if request_id is not None:
        selectionInfo["requestId"] = request_id

    os.mkdir(DATA_PATH + project_id + "/selections/" + selection_id)
    pythonModuleUtils.writeJsonFile(selectionInfoFilePath, selectionInfo)
    projects.update_project(project_id)
    return selectionInfo


def get_selections(project_id):
    # Get selections
    selections = []
    for selection_id in get_selection_ids(project_id):
        selections.append(get_selection(project_id, selection_id))
    return selections


def get_selection_ids(project_id):
    return os.listdir(DATA_PATH + project_id + "/selections/")


def selection_exist(project_id, selectionId):
    return os.path.exists(DATA_PATH + project_id + "/selections/" + selectionId)


def get_selection(project_id, selectionId):
    with open(
        DATA_PATH + project_id + "/selections/" + selectionId + "/info.json"
    ) as json_file:
        data = json.load(json_file)
        ret = {
            "id": data["id"],
            "name": data["name"],
            "filePath": data["filePath"],
            "creationDate": data["creationDate"],
            "updateDate": data["updateDate"],
            "nbSamples": len(data["samples"]),
        }

        # Add the request Id if it exist
        if "requestId" in data:
            ret["requestId"] = data["requestId"]

        return ret


def get_selection_id_list(project_id, selectionId):
    if not selection_exist(project_id, selectionId):
        raise Exception("Selection " + selectionId + " doesn't exist")

    with open(
        DATA_PATH + project_id + "/selections/" + selectionId + "/info.json"
    ) as json_file:
        data = json.load(json_file)
        return data["samples"]


# def getSelectionsSamples(project_id, selectionIds: list, intersection: bool) -> set:
#     if len(selectionIds) == 0:
#         return []

#     samples = set(get_selection_id_list(project_id, selectionIds[0]))
#     for selectionId in selectionIds[1:]:
#         if intersection:  # intersection of the selections samples
#             samples.intersection_update(
#                 get_selection_id_list(project_id, selectionId))

#             if len(samples) == 0:
#                 return []
#         else:  # Union of the model results samples
#             samples = samples.union(
#                 get_selection_id_list(project_id, selectionId))

#     return samples


def delete_selection(project_id, selection_id):
    pythonModuleUtils.deleteDir(DATA_PATH + project_id + "/selections/" + selection_id)
    projects.update_project(project_id)



import errno
import os
import argparse
from itertools import cycle
from shutil import get_terminal_size
from threading import Thread
from time import sleep
from threading import Timer
from termcolor import colored
from debiaiServer.config.init_config import DEBUG_COLOR, SUCCESS_COLOR


data_folder_path = "debiai_data"  # The path to the DebiAI data folder
DEFAULT_PORT = 3000  # default port

welcome_logo = """
                                                    █████████████            
                                                 ███████████████████         
                                               ███████████████████████       
                                             ██████████████████████████      
   █████████             ████        ███    ██████████   ██████   ███████    
    ███    ███            ███              ██████████     █████   ████████   
    ███     ███  ███████  ████████  ████   █████████   █   ████   ████████   
    ███     ███ ███   ███ ███   ███  ███   ████████   ███   ███   ████████   
    ███     ███ ████████  ███   ███  ███   ███████           ██   ████████   
    ███    ███  ███       ███   ███  ███   ███████   █████   ██   ████████   
   ██████████    ███████  ████████  █████  ██████   ███████   █   ███████    
                                            █████████████████████████████    
                                             ██████████████████████████      
                                            ██████████████████████████       
                                           █████████████████████████         
                                         ██████     █████████████            
                                       ██████                                
                                      ████                                   
"""


class Loader:
    def __init__(self, desc="Loading...", end="Done!", timeout=0.1):
        """
        A loader-like context manager

        Args:
            desc (str, optional): The loader's description. Defaults to "Loading...".
            end (str, optional): Final print. Defaults to "Done!".
            timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        """
        self.desc = desc
        self.end = end
        self.timeout = timeout

        self._thread = Thread(target=self._animate, daemon=True)
        self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        self.done = False

    def start(self):
        self._thread.start()
        return self

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break
            print(f"\r{self.desc} {c}", flush=True, end="")
            sleep(self.timeout)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.end}", flush=True)

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()


def create_folder(path=None):
    """Creates a folder at the specified path.

    Args:
        path (str, optional): The path to create the folder. If None, the current
            directory is used.

    Returns:
        str: The full path of the created folder.
    """

    if path is None:
        path = os.getcwd()

    try:
        # Check if the folder already exists
        if os.path.exists(path):
            # Check if the path is a folder
            if not os.path.isdir(path):
                raise OSError(
                    errno.ENOTDIR,
                    f"'{path}' is not a folder. Please provide a valid folder path.",
                )
            return os.path.abspath(path)  # Return the existing full path

        # If the folder does not exist, ask the user if they want to create it
        create_folder_answer = input(
            "The folder does not exist. Do you want to create it? (Y/n): "
        )
        if create_folder_answer == "Y" or create_folder_answer == "":
            # Create the folder
            os.makedirs(path)
            print(f"DebiAI data folder '{path}' created successfully.")
        else:
            print("Folder creation cancelled, exiting.")
            exit()

    except OSError as e:
        if e.errno == errno.EEXIST:
            # Folder already exists, handle gracefully
            print(f"Folder '{path}' already exists.")
            return os.path.abspath(path)  # Return the full existing path
        else:
            # Other error occurred, raise exception
            raise e

    return os.path.abspath(path)


def bash_info():
    print(
        colored("Usage: ", DEBUG_COLOR) + "debiai_gui [OPTIONS] COMMAND DATA_PATH\n\n"
        "\t"
        "Use the line below to run the app: \n\n"
        "\t\t"
        "$ debiai_gui"
        + colored(" start")
        + colored(" ./debiai_data\n\n")
        + colored("Options:\n", DEBUG_COLOR)
        + "\t"
        + colored("--version ", SUCCESS_COLOR)
        + "      Prints DebiAI version number.\n"
        + "\t"
        + colored("--port [NUMBER] ", SUCCESS_COLOR)
        + "Allows you to choose a port.\n"
        + colored("Commands:\n", DEBUG_COLOR)
        + "\t"
        + colored("start ", SUCCESS_COLOR)
        + "Starts the DebiAI GUI and open it in a web browser.\n\n"
        + "For more information visit: "
        + colored("https://debiai.irt-systemx.fr/ \n\n", None, attrs=["bold"])
    )


def run():
    # Parse the arguments
    parser = argparse.ArgumentParser(description="Run the DebiAI GUI")
    parser.add_argument("command", nargs="?", help="Start the DebiAI GUI server")
    parser.add_argument("path", nargs="?", help="The path to the DebiAI data folder.")
    parser.add_argument(
        "--port",
        type=int,
        default=DEFAULT_PORT,
        help="Port on which to run the DebiAI GUI, default is " + str(DEFAULT_PORT),
    )
    parser.add_argument(
        "--version", action="store_true", help="Prints the version number of DebiAI"
    )
    args = parser.parse_args()

    # Check the arguments
    if args.version:
        # Print the version number
        from debiaiServer.utils.utils import get_app_version

        version = get_app_version()
        print("DebiAI Version:" + colored(version, SUCCESS_COLOR))

    elif args.command == "start":
        global data_folder_path

        print(welcome_logo)
        print("   Welcome!\n")

        # Deal with the data folder path
        if args.path is None:
            # Set the default data folder path
            default_folder_name = "debiai_data"
            default_folder = os.path.join(os.getcwd(), default_folder_name)

            # Ask the user for the data folder path
            print("DebiAI requires a data folder to store the data.")
            data_folder_path_input = input(
                f"Enter the path of the DebiAI data folder ({default_folder}): "
            )
            if data_folder_path_input == "":
                data_folder_path_input = default_folder
            print("Data folder path input :", data_folder_path_input)
            data_folder_path = os.path.abspath(data_folder_path_input)
        else:
            # Use the provided data folder path
            data_folder_path = os.path.abspath(args.path)

        # Create the data folder
        data_folder_path_created = create_folder(data_folder_path)
        print("Data folder path:", data_folder_path_created, "\n")

        # Start the DebiAI GUI
        animation = Loader("Starting the DebiAI GUI", "")
        animation.start()
        start_server_with_browser(args.port, animation)

    else:
        # Print the usage information
        bash_info()


def start_server_with_browser(port, animation):
    from debiaiServer.websrv import start_server, open_browser

    Timer(1, lambda: open_browser(port)).start()
    animation.stop()
    start_server(port, reloader=False, is_dev=False)



import requests
import ujson as json

appUrl = "http://localhost:3000/"
test_algo_provider_name = "test_create_algo_provider"

algo_provider_list = []


def test_get_algorithms():
    global algo_provider_list
    url = appUrl + "app/algo-providers"
    resp = requests.get(url=url, headers={})
    assert resp.status_code == 200
    print(resp.text)
    algo_providers = json.loads(resp.text)
    assert type(algo_providers) is list

    for algo_provider in algo_providers:
        assert type(algo_provider) is dict
        assert "name" in algo_provider
        assert "url" in algo_provider
        assert "status" in algo_provider
        assert "algorithms" in algo_provider

        assert type(algo_provider["name"]) is str
        assert type(algo_provider["url"]) is str
        assert type(algo_provider["status"]) is bool
        assert type(algo_provider["algorithms"]) is list

        for algo in algo_provider["algorithms"]:
            assert type(algo) is dict
            assert "id" in algo
            assert "inputs" in algo
            assert "outputs" in algo

            assert type(algo["id"]) is str
            assert type(algo["inputs"]) is list
            assert type(algo["outputs"]) is list

            for input in algo["inputs"]:
                assert type(input) is dict
                assert "name" in input
                assert "type" in input

                assert type(input["name"]) is str
                assert type(input["type"]) is str

            for output in algo["outputs"]:
                assert type(output) is dict
                assert "name" in output
                assert "type" in output

                assert type(output["name"]) is str
                assert type(output["type"]) is str

    algo_provider_list = algo_providers


def test_add_algo_provider():
    url = appUrl + "app/algo-providers"
    data = {"name": test_algo_provider_name, "url": "http://localhost:4000"}
    resp = requests.post(url=url, headers={}, json=data)
    assert resp.status_code == 204

    # Test that it exists
    resp = requests.get(url=url, headers={})
    assert resp.status_code == 200
    print(resp.text)
    algo_providers = json.loads(resp.text)
    assert type(algo_providers) is list
    assert len(algo_providers) == len(algo_provider_list) + 1
    assert any(
        algo_provider["name"] == test_algo_provider_name
        for algo_provider in algo_providers
    )


def test_delete_algo_provider():
    url = appUrl + "app/algo-providers/" + test_algo_provider_name
    resp = requests.delete(url=url, headers={})
    assert resp.status_code == 204

    # Test that it was removed from the list
    url = appUrl + "app/algo-providers"
    resp = requests.get(url=url, headers={})
    assert resp.status_code == 200
    print(resp.text)
    algo_providers = json.loads(resp.text)
    assert type(algo_providers) is list
    assert len(algo_providers) == len(algo_provider_list)
    assert not any(
        algo_provider["name"] == test_algo_provider_name
        for algo_provider in algo_providers
    )



import debiaiServer.modules.dataProviders.webDataProvider.http.api as api
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)


def get_project_selections(url, project_id):
    try:
        selections = api.get_selections(url, project_id)

        if selections is None:
            print(f"Error: No selections found for project {project_id} on {url}")
            raise DataProviderException("No selections found", 404)

        debiai_selections = []
        for selection in selections:
            if "id" not in selection or selection["id"] is None:
                print(f"Error: No id for selection: {selection}")
                raise DataProviderException(
                    "An id is missing in the given selection", 400
                )

            selection_to_add = {
                "name": selection["name"] if "name" in selection else selection["id"],
                "id": selection["id"],
            }

            if "nbSamples" in selection:
                selection_to_add["nbSamples"] = selection["nbSamples"]
            if "creationDate" in selection:
                selection_to_add["creationDate"] = selection["creationDate"]
            if "updateDate" in selection:
                selection_to_add["updateDate"] = selection["updateDate"]
            if "metadata" not in selection:
                selection_to_add["metadata"] = None
            else:
                selection_to_add["metadata"] = selection["metadata"]

            debiai_selections.append(selection_to_add)
        return debiai_selections

    except DataProviderException:
        # The route may not be implemented in the data provider
        return []


def get_id_list_from_selection(url, cache, project_id, selection_id):
    id_list = cache.get_selection_id_list(project_id, selection_id)

    if id_list is None:
        id_list = api.get_selection_id(url, project_id, selection_id)
        cache.set_selection_id_list(project_id, selection_id, id_list)

    return id_list


def create_selection(url, project_id, name, id_list, request_id):
    data = {"idList": id_list, "name": name}

    if request_id is not None:
        data["request"]: request_id

    return api.post_selection(url, project_id, data)


def delete_selection(url, project_id, selection_id):
    return api.delete_selection(url, project_id, selection_id)



import pandas as pd
import numpy as np
import numpy.random as nr

from scipy.stats import pearsonr, spearmanr
from scipy.special import digamma
import scipy.spatial as ss
from scipy.spatial import cKDTree
from sklearn.neighbors import NearestNeighbors
from math import log, fabs, sqrt


#  === Correlation matrix ===
def pearsonCorrelation(data):
    """Computes the Pearson's coefficient for every pair of variables provided

    Parameters
    ----------
    data: list of lists where each list contains the observations of a single variable
        : Array of rows with the same sizes (discrete & continuous)

    Return
    ------
    result: correlation matrix along with the p-value of the significance test of the coefficients
            significance level legend: 3(***) -> p-value<0.01 -> The coefficient is significant at 99%
                                        2(**) -> p-value<0.05 ->The coefficient is significant at 95%
                                         1(*) -> p-value<0.1  -> The coefficient is significant at 90%

    """

    for i in range(len(data) - 1):
        assert len(data[i]) == len(
            data[i + 1]
        ), "The provided samples should have the same length"
    # transform the list of samples to dataframe
    df = pd.DataFrame(data).transpose()
    rho = df.corr()  # calculate the correlation matrix
    pval = df.corr(method=lambda x, y: pearsonr(x, y)[1]) - np.eye(
        *rho.shape
    )  # calculate the p-value
    # return the number of *
    p = pval.applymap(lambda x: (len([i for t in [0.01, 0.05, 0.1] if x <= t])))
    ret = rho.values.tolist()
    for i in range(rho.shape[0]):
        for j in range(rho.shape[1]):
            ret[i][j] = [float(rho[i][j]), float(p[i][j])]
    return (
        ret,
        200,
    )  # pearson correlation matrix with the significance of the coefficient


def spearmanCorrelation(data):
    """
    Computes the Spearman's coefficient for every pair of variables provided

        Parameters
        ----------
        data: list of lists where each list contains the observations of a single variable

        Return
        ------
        result: correlation matrix along with the p-value of the significance test of the coefficients
                significance level legend: 3(***) -> p-value<0.01 -> The coefficient is significant at 99%
                                            2(**) -> p-value<0.05 ->The coefficient is significant at 95%
                                             1(*) -> p-value<0.1  -> The coefficient is significant at 90%

    """
    for i in range(len(data) - 1):
        assert len(data[i]) == len(
            data[i + 1]
        ), "The provided samples should have the same length"
    # transform the list of samples to dataframe
    df = pd.DataFrame(data).transpose()
    rho = df.corr(method="spearman")  # calculate the correlation matrix
    pval = df.corr(method=lambda x, y: spearmanr(x, y)[1]) - np.eye(
        *rho.shape
    )  # calculate the p-value
    # return the number of significant level
    p = pval.applymap(lambda x: (len([i for t in [0.01, 0.05, 0.1] if x <= t])))
    result = rho.values.tolist()
    for i in range(rho.shape[0]):
        for j in range(rho.shape[1]):
            result[i][j] = [float(rho[i][j]), float(p[i][j])]
    return result, 200


# === Mutual Information ===
def entropy_discrete(x, base=2):
    """
    Computes the entropy of a discrete random variable

    Parameters:
    -----------

    x: List or array of one variable.
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats

    Return:
    -------

    Output: A float: The value of the entropy
    """
    _, count = np.unique(x, return_counts=True, axis=0)
    probability = count.astype(float) / len(x)
    # Removing the elements which have 0 probability/weight to avoid log(0)
    probability = probability[probability > 0.0]
    return np.sum(-1 * probability * np.log(probability)) / np.log(base)


def entropy_discrete_xy(x, y, base=2):
    """
    Computes the entropy of the joint distribution of two discrete random variables

    Parameters:
    -----------
    x,y : Two random variables samples of the same length
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats

    Returns:
    --------

    Output: The value of the entropy: a float

    """
    assert len(x) == len(y), "The two provided samples should be of the same length"
    xy = np.c_[x, y]
    # construction of point :
    # Example :
    # (x,y)
    # [[1. 2.]
    # [2. 4.]
    # [3. 5.]]
    return entropy_discrete(xy, base)


def discrete_mutual_information(x, y, base=2):
    """
    Computes the mutual information of two discrete random variables: x,y

    Parameters:
    -----------

    x,y: Two random variable samples of the same length
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats

    Returns:
    -------

    Output: The value of the mutual information
    """
    assert len(x) == len(y), "The two provided samples should be of the same length"
    return (
        entropy_discrete(x, base)
        + entropy_discrete(y, base)
        - entropy_discrete_xy(x, y, base)
    )


def continuous_mutual_information(x, y, k=1, base=2):
    """
    Computes the mutual information between two continuous random variables

    Parameters:
    -----------
    x,y: Data: lists or numpy arrays
    k: the number of neighbors to consider
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats

    Returns:
    --------

    Output: the mutual information
    """
    x, y = np.asarray(x), np.asarray(y)
    x, y = x.reshape(x.shape[0], -1), y.reshape(y.shape[0], -1)
    x = x + 1e-10 * np.random.random_sample(x.shape)
    y = y + 1e-10 * np.random.random_sample(y.shape)
    xy = np.c_[x, y]
    x_tree = cKDTree(x)
    y_tree = cKDTree(y)
    xy_tree = cKDTree(xy)
    # query with k=k+1 to return the nearest neighbor, not counting the data point itself
    dist, _ = xy_tree.query(xy, k=k + 1, p=np.inf)
    epsilon = dist[:, -1]

    # for each point, count the number of neighbors
    # whose distance in the x-subspace is strictly < epsilon
    # repeat for the y subspace
    n = len(x)
    nx = np.empty(n, dtype=np.int)
    ny = np.empty(n, dtype=np.int)
    for ii in range(n):
        if epsilon[ii] <= 1e-10:
            nx[ii] = len(x_tree.query_ball_point(x_tree.data[ii], r=1e-9, p=np.inf)) - 1
            ny[ii] = len(y_tree.query_ball_point(y_tree.data[ii], r=1e-9, p=np.inf)) - 1
        else:
            nx[ii] = (
                len(
                    x_tree.query_ball_point(
                        x_tree.data[ii], r=epsilon[ii] - 1e-9, p=np.inf
                    )
                )
                - 1
            )
            ny[ii] = (
                len(
                    y_tree.query_ball_point(
                        y_tree.data[ii], r=epsilon[ii] - 1e-9, p=np.inf
                    )
                )
                - 1
            )

    mi = (
        digamma(k) - np.mean(digamma(nx + 1) + digamma(ny + 1)) + digamma(n)
    ) / np.log(
        base
    )  # version (1) in krakow scientific paper

    return mi


def mixed_mutual_information(c, d, n_neighbors, base=10):
    """
    Compute mutual information between continuous and discrete variables.

    Parameters
    ----------
    c : ndarray, shape (n_samples,)
        Samples of a continuous random variable.
    d : ndarray, shape (n_samples,)
        Samples of a discrete random variable.
    n_neighbors : int
        Number of nearest neighbors to search for each point
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats


    Returns:
    --------
    Output: The mutual information value
    """
    n_samples = c.shape[0]
    c = c.reshape((-1, 1))

    radius = np.empty(n_samples)
    label_counts = np.empty(n_samples)
    k_all = np.empty(n_samples)
    nn = NearestNeighbors()
    for label in np.unique(d):
        mask = d == label
        count = np.sum(mask)
        if count > 1:
            k = min(n_neighbors, count - 1)
            nn.set_params(n_neighbors=k)
            nn.fit(c[mask])
            r = nn.kneighbors()[0]
            # print(r)
            radius[mask] = np.nextafter(r[:, -1], 0)
            # print(radius)
            k_all[mask] = k
        label_counts[mask] = count

    # Ignore points with unique labels.
    mask = label_counts > 1
    n_samples = np.sum(mask)
    label_counts = label_counts[mask]
    k_all = k_all[mask]
    c = c[mask]
    radius = radius[mask]

    nn.set_params(algorithm="kd_tree")
    nn.fit(c)
    ind = nn.radius_neighbors(radius=radius, return_distance=False)
    m_all = np.array([i.size for i in ind])

    mi = (
        digamma(n_samples)
        + np.mean(digamma(k_all))
        - np.mean(digamma(label_counts))
        - np.mean(digamma(m_all + 1))
    )

    return mi / log(base)


def normalise_function(normalise, mutual_information, entropy_X, entropy_Y):
    """
    normalize the mutual information coefficient
    Parameters:
    -----------
    normalize: the choice of normalize function : takes either 'max' or 'min' or 'square root' or 'mean' or 'none'
    mutual_information: mutual information coefficient
    entropy_X: the entropy of the first variable
    entropy_Y: the entropy of the second variable
    Returns:
    -------
    Output:
    The value of the normalized mutual information coefficient
    """
    if normalise == "none":
        ratio = 1
    elif normalise == "max":
        ratio = max(entropy_X, entropy_Y)
    elif normalise == "min":
        ratio = min(entropy_X, entropy_Y)
    elif normalise == "square root":
        ratio = sqrt(np.abs(entropy_X * entropy_Y))
    elif normalise == "mean":
        ratio = (entropy_X + entropy_Y) / 2
    else:
        raise NotImplementedError(
            "Variable 'normalise' takes only 'max' or 'min' or 'square root' or 'mean' or 'none'"
        )

    return mutual_information / ratio


def continuous_iterate_function(list_continuous, k=3, base=3, normalise="none"):
    """
      Parameters:
     -----------
    list_continuous: list of list of the continuous variables
    k: the number of neighbors to consider
     base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats
     normalise: the choice of normalize function : takes either 'max' or 'min' or 'square root' or 'mean' or 'none'

     Returns:
     -------
     Output:
     an array of the mutual information between the continuous variables
    """
    continuous = np.eye(len(list_continuous), len(list_continuous)).tolist()
    for i in range(len(continuous)):
        for j in range(i, len(continuous)):
            continuous[i][j] = continuous[j][i] = normalise_function(
                normalise,
                continuous_mutual_information(
                    list_continuous[i], list_continuous[j], k=k, base=base
                ),
                continuous_mutual_information(
                    list_continuous[i], list_continuous[i], k=k, base=base
                ),
                continuous_mutual_information(
                    list_continuous[j], list_continuous[j], k=k, base=base
                ),
            )
    return continuous


def discrete_iterate_function(list_discrete, base=10, normalise="none"):
    """
     Parameters:
     -----------
    list_discrete: list of list of the discrete variables
    k: the number of neighbors to consider
     base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats
     normalise: the choice of normalize function : takes either 'max' or 'min' or 'square root' or 'mean' or 'none'

     Returns:
     -------
     Output:
     an array of the mutual information between the discrete variables
    """
    discrete = np.eye(len(list_discrete), len(list_discrete)).tolist()
    for i in range(len(discrete)):
        for j in range(i, len(discrete)):
            discrete[i][j] = discrete[j][i] = normalise_function(
                normalise,
                discrete_mutual_information(
                    list_discrete[i], list_discrete[j], base=base
                ),
                entropy_discrete(list_discrete[i], base=base),
                entropy_discrete(list_discrete[j], base=base),
            )
    return discrete


def mixed_iterate_function(
    list_continuous, list_discrete, base=10, k=3, normalise="none"
):
    """
     Parameters:
    -----------
    list_continuous: list of list of the continuous variables
    list_discrete: list of list of the discrete variables
    k: the number of neighbors to consider
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats
    normalise: the choice of normalize function : takes either 'max' or 'min' or 'square root' or 'mean' or 'none'

    Returns:
    -------
    Output:
    an array of the mutual information between the all variables: the continuous and the discrete ones
    """
    mixed = np.eye(len(list_continuous), len(list_discrete)).tolist()
    for i in range(len(list_continuous)):
        for j in range(len(list_discrete)):
            mixed[i][j] = normalise_function(
                normalise,
                mixed_mutual_information(
                    np.array(list_continuous[i]),
                    np.array(list_discrete[j]),
                    n_neighbors=k,
                    base=base,
                ),
                continuous_mutual_information(
                    list_continuous[i], list_continuous[i], k=k, base=base
                ),
                entropy_discrete(list_discrete[j], base=base),
            )
    mixed = pd.DataFrame(mixed)
    continuous = continuous_iterate_function(
        list_continuous, k=k, base=base, normalise=normalise
    )
    continuous = pd.DataFrame(continuous)
    discrete = discrete_iterate_function(list_discrete, base=base, normalise=normalise)
    discrete = pd.DataFrame(discrete)
    part1 = np.concatenate((continuous, mixed), axis=1)
    part2 = np.concatenate((mixed.T, discrete), axis=1)
    result = np.concatenate((part1, part2), axis=0)
    return result.tolist()


# @utils.traceLogLight
def mutualInformation(data):
    """
     the global matrix of mutual information
    Parameters:
    -----------
    list_continuous: list of list of the continuous variables, if there is no
    continuous variables, please send an empty list of list [[]]
    list_discrete: list of list of the discrete variables,if there is no discrete
    variables, please send an empty list of list [[]]
    k: the number of neighbors to consider
    base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats
    normalise: the choice of normalize function : takes either 'max' or 'min'
    or 'square root' or 'mean' or 'none'

    Returns:
    -------
    Output:
    an array of the mutual information between the variables: the continuous
    ones only, the discrete ones only or the continuous and the discrete ones
    """
    k = data["k"]
    list_continuous = data["list_continuous"]
    list_discrete = data["list_discrete"]

    if k >= len(list_continuous) + len(list_discrete):
        return "k must be lower than the number of variables", 403

    if "base" in data:
        base = data["base"]
    else:
        base = 2

    if "normalise" in data:
        normalise = data["normalise"]
        acceptedNormalise = ["max", "min", "square root", "mean", "none"]

        if normalise not in acceptedNormalise:
            return (
                "normalise need to be either 'max' or 'min' or 'square root' or 'mean' or 'none'",
                403,
            )
    else:
        normalise = "max"

    # Calculate mutual information between features
    if list_continuous != [[]] and list_discrete == [[]]:
        return continuous_iterate_function(
            list_continuous, k=k, base=base, normalise=normalise
        )
    elif list_continuous == [[]] and list_discrete != [[]]:
        return discrete_iterate_function(list_discrete, base=base, normalise=normalise)
    elif list_continuous != [[]] and list_discrete != [[]]:
        return mixed_iterate_function(
            list_continuous, list_discrete, base=base, k=k, normalise=normalise
        )
    else:
        return "The lists are empty", 403


# === Mutual Information higher dimension ===
def averageDigamma(points, dvec):
    """
    This part finds number of neighbors in some radius in the marginal space

    Parameters:
    ----------

    points: the data observed
    dvec: A distance vector between points

    Returns:
    --------
    Output: expectation value of <psi(nx)>

    """

    N = len(points)
    tree = ss.cKDTree(points)
    avg = 0.0
    for i in range(N):
        dist = dvec[i]
        # subtlety, we don't include the boundary point,
        # but we are implicitly adding 1 to kraskov definition because center point is included
        num_points = len(tree.query_ball_point(points[i], dist - 1e-15, p=float("inf")))
        avg += digamma(num_points) / N
    return avg


# @utils.traceLogLight
def higherDimensionMutualInformation(data):
    """
    This function calculates the mutual information between several continuous
    variables (3, 4 variables). It takes as input a list of lists [[variable1],
    [variable2], [variable3], ...], the number K, and the base, either 2 or 10
    (the unit of information is respectively bits or nats).

    It returns the mutual information between the different variables.

    Regarding its representation in the tool, we can simply create a small window
    where we select the variables, K, and the base, and display only the result.
    Alternatively, in your 3D plot window, we can add the result of the mutual
    information of the 3 variables below the graph.

    The mutual information estimator by Kraskov et al.
    ith row of X represents ith dimension of the data, e.g. X = [[1.0,3.0,3.0],[0.1,1.2,5.4]],
    if X has two dimensions and we have three samples
    Parameters:
    ----------
        X: list of list of the variables : it could take more than 2 variables
        k: the number of neighbors to consider
        base: The base in which the entropy value is represented, i.e 2 for bits, 10 for nats

    Returns:
    --------
        Output: the mutual information between the variables
    """

    X = data["X"]
    k = data["k"]
    if k >= len(X):
        return "k must be < to len(X)", 403

    if "base" in data:
        base = data["base"]
    else:
        base = 2

    # adding small noise to X, e.g., x<-X+noise
    x = []
    for i in range(len(X)):
        tem = []
        for j in range(len(X[i])):
            tem.append([X[i][j] + 1e-10 * nr.rand(1)[0]])
        x.append(tem)

    points = []
    for j in range(len(x[0])):
        tem = []
        for i in range(len(x)):
            tem.append(x[i][j][0])
        points.append(tem)
    tree = ss.cKDTree(points)
    dvec = []
    for i in range(len(x)):
        dvec.append([])
    for point in points:
        # Find k-nearest neighbors in joint space, p=inf means max norm
        knn = tree.query(point, k + 1, p=float("inf"))
        points_knn = []
        for i in range(len(x)):
            dvec[i].append(float("-inf"))
            points_knn.append([])
        for j in range(k + 1):
            for i in range(len(x)):
                points_knn[i].append(points[knn[1][j]][i])

        # Find distances to k-nearest neighbors in each marginal space
        for i in range(k + 1):
            for j in range(len(x)):
                if dvec[j][-1] < fabs(points_knn[j][i] - points_knn[j][0]):
                    dvec[j][-1] = fabs(points_knn[j][i] - points_knn[j][0])

    ret = 0.0
    for i in range(len(x)):
        ret -= averageDigamma(x[i], dvec[i])
    ret += (
        digamma(k)
        - (float(len(x)) - 1.0) / float(k)
        + (float(len(x)) - 1.0) * digamma(len(x[0]))
    )
    if base == 2:
        mul = 1 / log(2)  # scaling factor from nats to bits
        ret *= mul
    return ret

    pass


# === Mutual Information matrix & Mutual Information higher dimension ===
def mutualAndHigherInformation(data):
    """
    calculate the mutual information estimator by Kraskov et al.
    and the global matrix of mutual information
    """

    k = data["k"]
    list_continuous = data["list_continuous"]
    list_discrete = data["list_discrete"]
    columns = list_continuous + list_discrete
    if "base" in data:
        base = data["base"]
    else:
        base = 2

    if "normalise" in data:
        normalise = data["normalise"]
    else:
        normalise = "max"

    if k >= len(columns):
        return "k need to be < len(X)", 403

    # higherDimensionMutualInformation
    print("higherDimensionMutualInformation")
    hdmi = higherDimensionMutualInformation(
        {"k": k, "base": base, "X": list_continuous + list_discrete}
    )

    # mutualInformation
    print("mutualInformation")
    mi = mutualInformation(
        {
            "k": k,
            "base": base,
            "list_continuous": list_continuous,
            "list_discrete": list_discrete,
            "normalise": normalise,
        }
    )

    return {"higherDimensionMutualInformation": hdmi, "mutualInformation": mi}, 200



#############################################################################
# Imports
#############################################################################
import debiaiServer.modules.dataProviders.dataProviderManager as data_provider_manager
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)

#############################################################################
# Data Management
#############################################################################


def get_data(dataProviderId, projectId, data):
    # return a project data from a list of ids
    sampleIds = data["sampleIds"]
    analysis = data["analysis"]

    try:
        # Find the data provider
        data_provider = data_provider_manager.get_single_data_provider(dataProviderId)

        # Ask for the data
        samples = data_provider.get_samples(projectId, analysis, sampleIds)

        if samples is not None:
            return {
                "data": samples,
                "dataMap": True,
            }, 200

        return (
            "Can't find samples for project "
            + projectId
            + " on data provider : "
            + dataProviderId,
            404,
        )
    except DataProviderException as e:
        return e.message, e.status_code



from debiaiServer.modules.dataProviders.pythonDataProvider.dataUtils import (
    pythonModuleUtils,
    tree,
    hash,
)

DATA_TYPES = pythonModuleUtils.DATA_TYPES

# ID list


def get_all_samples_id_list(project_id, _from=None, _to=None):
    """
    Return a list of all samples id in a project
    """
    # Get the hashmap
    hashmap = hash.getHashmap(project_id)

    # Get all samples
    samples = list(hashmap.keys())

    # In case of streaming purpose
    if _from is not None and _to is not None:
        samples = samples[_from : _to + 1]  # noqa

    return samples


# Get data
def get_data_from_sample_id_list(project_id, id_list):
    # Get path of the samples from the hashmap
    sample_path = hash.getPathFromHashList(project_id, id_list)
    data = {}

    # We age going through each samples individually because of a bug
    # (the data aren't aligned with the requested samples id)
    # Because of this bug, we are slowing down the process
    # TODO : fix this bug
    for i in range(len(id_list)):
        # Get tree from samples
        samples_tree = tree.getBlockTreeFromSamples(project_id, [sample_path[i]])

        # Convert tree to array
        data_array = _tree_to_array(samples_tree)

        # Convert array to dict
        data[id_list[i]] = data_array[0]

    return data


def _tree_to_array(tree):
    data_array = []
    for block in tree:
        data_array += _block_to_array_recur(block)
    return data_array


def _get_block_values(block):
    # Adding the block name into the values
    values = [block["name"]]

    # store all key-values into an array
    for data_type in DATA_TYPES:
        if data_type in block:
            for key in range(len(block[data_type])):
                values.append(block[data_type][key])

    return values


def _block_to_array_recur(block):
    # Getting bloc values
    values = _get_block_values(block)
    if "childrenInfoList" not in block or len(block["childrenInfoList"]) == 0:
        return [values]

    else:
        # Getting all child values
        child_values = []
        for child_block in block["childrenInfoList"]:
            child_values += _block_to_array_recur(child_block)
            # child_values.append(_block_to_array_recur(child_block))

        # Child values : [[1,2,3], [4,5,6], [7,8,9]]
        # values : [10, 11, 12]
        # Goal: [[10, 11, 12, 1, 2, 3], [10, 11, 12, 4, 5, 6], [10, 11, 12, 7, 8, 9]]

        # Adding the block name into the values
        ret = [None] * len(child_values)

        # Adding the block values to the children values
        for i in range(len(child_values)):
            ret[i] = values + child_values[i]

        return ret


# def projectSamplesGenerator(projectId):
#     """
#     Generator used to iterate over all samples in a project.
#     Used by the 'createSelectionFromRequest' method
#     """

#     # Get the project block structure
#     projectBlockStructure = projects.get_project_block_level_info(projectId)
#     sampleLevel = len(projectBlockStructure) - 1

#     rootBlocks = utils.listDir(DATA_PATH + projectId + "/blocks/")
#     for rootBlock in rootBlocks:
#         path = DATA_PATH + projectId + "/blocks/" + rootBlock + "/"
#         yield from yieldSample(path,  0, [], sampleLevel, projectBlockStructure)
#     print("end")


# def yieldSample(path, level, sampleInfo, sampleLevel, blockLevelInfo):
#     # TODO : optimizations : add in parameters the block that we need to open
#     blockInfo = utils.readJsonFile(path + "info.json")
#     sampleInfo.append(getBlockInfo(blockLevelInfo[level], blockInfo))

#     if level == sampleLevel:
#         # merge the dict into one
#         yield {k: v for x in sampleInfo for k, v in x.items()}, blockInfo["id"]
#     else:
#         childrenBlockNames = utils.listDir(path)
#         for name in childrenBlockNames:
#             yield from yieldSample(path + name + "/",
#                                    level + 1, sampleInfo, sampleLevel, blockLevelInfo)
#             del sampleInfo[-1]



import psutil
import requests
import connexion
import webbrowser
from flask_cors import CORS
from waitress import serve
from termcolor import colored
from debiaiServer.init import init
from debiaiServer.utils.utils import get_app_version
from debiaiServer.config.init_config import DEBUG_COLOR
from flask import send_from_directory, request, Response


DEV_FRONTEND_URL = "http://localhost:8080/"
app = connexion.App(__name__)
app.add_api("swagger.yaml", strict_validation=True)
CORS(app.app)


def send_frontend(path, is_dev):
    if path == "/":
        path = "index.html"

    # If production, use the index.html from the dist folder
    if not is_dev:
        return send_from_directory("dist", path)

    # In development, redirect to the DEV_FRONTEND_URL
    else:
        if request.method == "GET":
            try:
                resp = requests.get(f"{DEV_FRONTEND_URL}{path}")
                excluded_headers = [
                    "content-encoding",
                    "content-length",
                    "transfer-encoding",
                    "connection",
                ]
                headers = [
                    (name, value)
                    for (name, value) in resp.raw.headers.items()
                    if name.lower() not in excluded_headers
                ]
                response = Response(resp.content, resp.status_code, headers)
                return response
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
                return (
                    "You are in a development environment and the DebiAI frontend"
                    + "is not available at the url : "
                    + DEV_FRONTEND_URL,
                    503,
                )
        else:
            print("Unexpected request method")


def create_app(is_dev):
    # For serving the dashboard
    @app.route("/")
    def send_index():
        return send_frontend("/", is_dev)

    # For serving the dashboard assets
    @app.route("/<path:path>")
    def send_supporting_elements(path):
        return send_frontend(path, is_dev)

    return app


def is_browser_open():
    """Check if a browser process is running."""
    browser_keywords = ["chrome", "firefox", "safari", "edge", "opera"]
    for proc in psutil.process_iter(["pid", "name"]):
        for keyword in browser_keywords:
            if keyword in proc.info["name"].lower():
                return True
    return False


def open_browser(port):
    url = f"http://localhost:{port}"
    if is_browser_open():
        webbrowser.open_new_tab(url)
    else:
        webbrowser.open(url)


def start_server(port, reloader=True, is_dev=True):
    # Run DebiAI init
    print(
        "================= DebiAI " + get_app_version() + " ====================",
        flush=True,
    )
    init()
    print("======================== RUN =======================", flush=True)
    print(
        "   DebiAI is available at "
        + colored("http://localhost:" + str(port), DEBUG_COLOR),
        flush=True,
    )
    app = create_app(is_dev)
    if is_dev:
        # Use flask server for development
        app.run(port, debug=True, host="0.0.0.0", use_reloader=reloader)
    else:
        # Use waitress for production
        serve(app, host="0.0.0.0", port=port, threads=6)

    # Print goodbye message
    print("\n====================================================", flush=True)
    print("\nGoodbye!")
    print("Thank you for using DebiAI.\n")


if __name__ == "__main__":
    start_server(port=3000, reloader=False, is_dev=True)



import ujson as json
import os
import posixpath
import unicodedata
import string
import shutil
import time
from debiaiServer.debiai_gui_utils import data_folder_path

DATA_PATH = data_folder_path + "/pythonDataProvider/"

DATA_TYPES = ["groundTruth", "contexts", "inputs", "others"]


# Init, called at the server start
def init():
    # Create the folder if it does not exist
    if not os.path.exists(DATA_PATH):
        os.mkdir(DATA_PATH)


# File name verifications
def clean_filename(filename):
    # replace spaces
    filename = filename.replace(" ", "_")

    # keep only valid ascii chars
    cleaned_filename = (
        unicodedata.normalize("NFKD", filename).encode("ASCII", "ignore").decode()
    )

    # keep only whitelisted chars
    whitelist = "_-() %s%s" % (string.ascii_letters, string.digits)
    char_limit = 255

    cleaned_filename = "".join(c for c in cleaned_filename if c in whitelist)
    return cleaned_filename[:char_limit]


def is_filename_clean(filename):
    cleanFilename = "".join(i for i in filename if i not in "\/:*?<>|")  # noqa
    return filename == cleanFilename


def is_secure_path(path):
    path = posixpath.normpath(path)
    return not path.startswith(("/", "../"))


# directories and file Manipulation
def fileExist(path):
    return os.path.isfile(path)


def listDir(path):
    # List the directories only
    return [
        name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name))
    ]


def deleteFile(filePath):
    os.remove(filePath)


def deleteDir(dirPath):
    deleteFiles = []
    deleteDirs = []
    for root, dirs, files in os.walk(dirPath):
        for f in files:
            deleteFiles.append(os.path.join(root, f))
        for d in dirs:
            deleteDirs.append(os.path.join(root, d))
    for f in deleteFiles:
        os.remove(f)
    for d in deleteDirs:
        os.rmdir(d)
    os.rmdir(dirPath)


def copyDir(src, dest):
    shutil.copytree(src, dest)


# Json files
def readJsonFile(path):
    with open(path, "r") as jsonFile:
        return json.load(jsonFile)


def writeJsonFile(path, obj):
    with open(path, "w") as outfile:
        json.dump(obj, outfile)


def updateJsonFile(path, key, data):
    with open(path, "r") as jsonFile:
        d = json.load(jsonFile)

    d[key] = data

    with open(path, "w") as jsonFile:
        json.dump(d, jsonFile)
    return d


def addToJsonFIle(path, dictToAdd: dict):
    with open(path, "r") as jsonFile:
        d = json.load(jsonFile)

    d = {**d, **dictToAdd}
    with open(path, "w") as jsonFile:
        json.dump(d, jsonFile)
    return d


# Date
def timeNow():
    return time.time() * 1000



from termcolor import colored

from debiaiServer.config.init_config import (
    get_config,
    DEBUG_COLOR,
    ERROR_COLOR,
    SUCCESS_COLOR,
)
from debiaiServer.modules.dataProviders.webDataProvider.WebDataProvider import (
    WebDataProvider,
)
from debiaiServer.modules.dataProviders.pythonDataProvider.PythonDataProvider import (
    PythonDataProvider,
    PYTHON_DATA_PROVIDER_ID,
)
from debiaiServer.modules.dataProviders.DataProviderException import (
    DataProviderException,
)

data_providers_list = []
python_data_provider_disabled = True


def setup_data_providers():
    global python_data_provider_disabled
    print("================== DATA PROVIDERS ==================")
    config = get_config()
    web_data_provider_config = config["WEB_DATA_PROVIDERS"]
    python_module_data_provider_config = config["INTEGRATED_DATA_PROVIDER"]

    keys = list(web_data_provider_config.keys())
    values = list(web_data_provider_config.values())

    # Web Data Providers
    for i in range(len(web_data_provider_config)):
        name = keys[i]
        url = values[i]

        # Remove trailing slash
        if url[-1] == "/":
            url = url[:-1]

        print(
            " - Adding external data Provider "
            + colored(name, DEBUG_COLOR)
            + " ("
            + url
            + ")"
        )
        try:
            data_provider = WebDataProvider(url, name)
            add(data_provider)

            if data_provider.is_alive():
                print(colored("   [SUCCESS]", SUCCESS_COLOR) + " Data Provider ready")
            else:
                raise DataProviderException()
        except DataProviderException:
            print(
                colored("   [ERROR]", ERROR_COLOR)
                + " : Data Provider "
                + colored(name, ERROR_COLOR)
                + " is not accessible"
            )
    # Python Data Providers
    if python_module_data_provider_config["enabled"]:
        print(" - Adding Python Module data Provider")
        add(PythonDataProvider())
        python_data_provider_disabled = False

    if len(data_providers_list) == 0:
        print("   No data providers configured")


def data_provider_exists(name):
    for d in data_providers_list:
        if d.name == name:
            return True
    return False


def is_valid_name(name):
    # /, &, | are not allowed in data provider names
    if (
        "/" in name
        or "&" in name
        or "|" in name
        or len(name) == 0
        or len(name) > 50
        or name[0] == " "
        or name[-1] == " "
    ):
        return False

    return True


def add(data_provider):
    data_providers_list.append(data_provider)
    return


def get_data_provider_list():
    return data_providers_list


def get_single_data_provider(name):
    # Check if the data provider is not disabled
    if name == PYTHON_DATA_PROVIDER_ID and python_data_provider_disabled:
        raise DataProviderException("Python module data provider is disabled", 403)

    # Return the data provider with the given name
    for d in data_providers_list:
        if d.name == name:
            return d

    raise DataProviderException("Data provider not found", 404)


def delete(name):
    for d in data_providers_list:
        if d.name == name:
            if d.type == "Python module Data Provider":
                raise DataProviderException(
                    "Python module data provider cannot be deleted", 403
                )

            data_providers_list.remove(d)
            return



import os
import json
import debiaiServer.utils.utils as utils
import uuid
from debiaiServer.debiai_gui_utils import data_folder_path


LAYOUTS_PATH = data_folder_path + "/layouts.json"


# Layouts file structure
# [
#   {
#       "id": ""
#       "name": ""
#       "description": "",
#       "projectId": ""
#       "dataProviderId": ""
#       "creationDate": 0
#       "layout": [
#          # Widget position
#          {
#            "widgetKey": "parallelCoordinate",
#            "x": 0,
#            "y": 0,
#            "width": 0,
#            "height": 0,
#            "config": {}, # Widget config (optional)
#            "name": "", # Name given to the widget (optional)
#            "localFilters" : [{}],
#          },
#       ],
#       "selectedColorColumn": "col", # (optional)
#   },
#   ...
# ]


def setup_layouts():
    # Create the file if it does not exist
    if not os.path.exists(LAYOUTS_PATH):
        with open(LAYOUTS_PATH, "w") as json_file:
            json.dump([], json_file)
            json_file.flush()
            json_file.close()


def get_layouts():
    # Return the layouts list
    try:
        if os.path.getsize(LAYOUTS_PATH) == 0:
            # File exists but is empty, reset to default layouts
            print("Layouts file is empty, resetting to default layouts.")
            _save_layouts([])
            return []
        with open(LAYOUTS_PATH) as json_file:
            layouts = json.load(json_file)
            return layouts

    except FileNotFoundError:
        # File does not exist, create it with default layouts
        print("Layouts file not found, creating with default layouts.")
        setup_layouts()
        return []

    except json.decoder.JSONDecodeError as e:
        # Print the error and reset the file with default layouts
        print("Error while reading the layouts file:", e)
        print("The file will be reset to default layouts.")
        _save_layouts([])
        return []


def add_layout(data):
    # project_id, data_provider_id, conf_description, conf_name, conf
    # Add a new widget layout
    # Generate id
    id = str(uuid.uuid1())

    layout_to_add = []

    for widget in data["layout"]:
        widget_position = {
            "x": widget["x"],
            "y": widget["y"],
            "width": widget["width"],
            "height": widget["height"],
            "widgetKey": widget["widgetKey"],
        }

        keys = ["config", "name", "localFilters"]

        for key in keys:
            if key in widget:
                widget_position[key] = widget[key]

        layout_to_add.append(widget_position)

    file_to_add = {
        "id": id,
        "name": data["name"],
        "description": data["description"],
        "projectId": data["projectId"],
        "dataProviderId": data["dataProviderId"],
        "creationDate": utils.timeNow(),
        "layout": layout_to_add,
        "lastLayoutSaved": False,
    }

    if "selectedColorColumn" in data:
        file_to_add["selectedColorColumn"] = data["selectedColorColumn"]

    layouts = get_layouts()

    # Check if their is already a "last saved" layout
    if "lastLayoutSaved" in data and data["lastLayoutSaved"]:
        file_to_add["lastLayoutSaved"] = True

        for layout in layouts:
            if (
                layout["projectId"] == data["projectId"]
                and layout["dataProviderId"] == data["dataProviderId"]
                and "lastLayoutSaved" in layout
                and layout["lastLayoutSaved"]
            ):
                # Remove the "last saved" layout
                layouts.remove(layout)

    # Save layout
    layouts.append(file_to_add)
    _save_layouts(layouts)


def delete_layout(id):
    # Delete the widget layout by its name
    layouts = get_layouts()

    for layout in layouts:
        if layout["id"] == id:
            layouts.remove(layout)

    _save_layouts(layouts)


def _save_layouts(layouts, _second_try=False):
    # Update the json file
    try:
        with open(LAYOUTS_PATH, "w") as json_file:
            json.dump(layouts, json_file, ensure_ascii=False, indent=4)
            # Ensure data is written before closing the file
            json_file.flush()
            json_file.close()
    except FileNotFoundError:
        if not _second_try:
            setup_layouts()
            _save_layouts(layouts, _second_try=True)
        else:
            print("Error while saving the layouts file")



import hashlib
import ujson as json

from debiaiServer.modules.dataProviders.pythonDataProvider.dataUtils import (
    pythonModuleUtils,
)

DATA_PATH = pythonModuleUtils.DATA_PATH


def hash(text: str):
    return hashlib.sha256(text.encode("utf-8")).hexdigest()


# hash
def __createProjectHashMap(projectId, blockPath, hashmap, sampleLevel, currentLevel):
    blockPath += "/"
    if currentLevel == sampleLevel:
        # We are at the sample level, we can fill the hashmap
        sampleHash = hash(blockPath)
        hashmap[sampleHash] = blockPath

        # Update the sample
        pythonModuleUtils.updateJsonFile(
            DATA_PATH + projectId + "/blocks/" + blockPath + "info.json",
            "id",
            sampleHash,
        )
        return

    for children in pythonModuleUtils.listDir(
        DATA_PATH + projectId + "/blocks/" + blockPath
    ):
        __createProjectHashMap(
            projectId, blockPath + children, hashmap, sampleLevel, currentLevel + 1
        )


def addToSampleHashmap(projectId, hashMap):
    with open(DATA_PATH + projectId + "/samplesHashmap.json") as json_file:
        existingHm = json.load(json_file)

    existingHm.update(hashMap)

    pythonModuleUtils.writeJsonFile(
        DATA_PATH + projectId + "/samplesHashmap.json", existingHm
    )


def getHashmap(projectId):
    with open(DATA_PATH + projectId + "/samplesHashmap.json") as json_file:
        existingHm = json.load(json_file)

    return existingHm


def getPathFromHashList(projectId, hashArray):
    hm = getHashmap(projectId)
    ret = []
    for hash in hashArray:
        ret.append(hm[hash])
    return ret
